@@ -16,9 +16,9 @@ from epsilon.extime import Time
 
 from axiom.slotmachine import Attribute as inmemory
 
-from axiom.errors import NoCrossStoreReferences
+from axiom.errors import NoCrossStoreReferences, BrokenReference
 
-from axiom.iaxiom import IComparison, IOrdering
+from axiom.iaxiom import IComparison, IOrdering, IColumn, IQuery
 
 USING_APSW = False
 
@@ -26,45 +26,18 @@ _NEEDS_FETCH = object()         # token indicating that a value was not found
 
 __metaclass__ = type
 
-class Comparable:
+
+class _ComparisonOperatorMuxer:
     """
-    Helper for a thing that can be compared like an SQLAttribute (or is in fact
-    an SQLAttribute).  Requires that 'self' have 'type' (Item-subclass) and
-    'columnName' (str) attributes, as well as an 'infilter' method in the
-    spirit of SQLAttribute, documented below.
+    Collapse comparison operations into calls to a single method with varying
+    arguments.
     """
-
-    def compare(self, other, sqlop):
-        # interim: maybe we want objects later?  right now strings should be fine
-        a = []
-        tables = [self.type]
-        if isinstance(other, Comparable):
-            return TwoAttributeComparison(self, sqlop, other)
-        elif other is None:
-            if sqlop == '=':
-                negate = False
-            elif sqlop == '!=':
-                negate = True
-            else:
-                raise TypeError(
-                    "None/NULL does not work with %s comparison" % (sqlop,))
-            return NullComparison(self, negate)
-        else:
-            # convert to constant usable in the database
-            return AttributeValueComparison(self, sqlop, other)
-
-    def oneOf(self, seq, negate=False):
+    def compare(self, other, op):
         """
-        Choose items whose attributes are in a fixed set.
-
-        X.oneOf([1, 2, 3])
-
-        Implemented with the SQL 'in' statement.
+        Override this in a subclass.
         """
-        return SequenceComparison(self, list(seq), negate)
+        raise NotImplementedError()
 
-    def notOneOf(self, seq):
-        return self.oneOf(seq, negate=True)
 
     def __eq__(self, other):
         return self.compare(other, '=')
@@ -90,24 +63,53 @@ class Comparable:
         return self.compare(other, '<=')
 
 
-    _likeOperators = ('LIKE', 'NOT LIKE')
-    def _like(self, negate, *others):
-        if not others:
-            raise ValueError, 'Must pass at least one expression to _like'
+def compare(left, right, op):
+    # interim: maybe we want objects later?  right now strings should be fine
+    a = []
+    if IColumn.providedBy(right):
+        return TwoAttributeComparison(left, op, right)
+    elif right is None:
+        if op == '=':
+            negate = False
+        elif op == '!=':
+            negate = True
+        else:
+            raise TypeError(
+                "None/NULL does not work with %s comparison" % (op,))
+        return NullComparison(left, negate)
+    else:
+        # convert to constant usable in the database
+        return AttributeValueComparison(left, op, right)
 
+
+
+class _MatchingOperationMuxer:
+    """
+    Collapse string matching operations into calls to a single method with
+    varying arguments.
+    """
+    def _like(self, negate, firstOther, *others):
+        others = (firstOther,) + others
         likeParts = []
 
+        allValues = True
         for other in others:
-            if isinstance(other, Comparable):
+            if IColumn.providedBy(other):
                 likeParts.append(LikeColumn(other))
+                allValues = False
             elif other is None:
                 # LIKE NULL is a silly condition, but it's allowed.
                 likeParts.append(LikeNull())
+                allValues = False
             else:
                 likeParts.append(LikeValue(other))
 
+        if allValues:
+            likeParts = [LikeValue(''.join(others))]
+
         return LikeComparison(self, negate, likeParts)
 
+
     def like(self, *others):
         return self._like(False, *others)
 
@@ -124,21 +126,58 @@ class Comparable:
         return self._like(False, '%', other)
 
 
-    # XXX TODO: improve error reporting
 
+_ASC = 'ASC'
+_DESC = 'DESC'
+
+class _OrderingMixin:
+    """
+    Provide the C{ascending} and C{descending} attributes to specify sort
+    direction.
+    """
     def _asc(self):
-        return SimpleOrdering(self, ASC)
+        return SimpleOrdering(self, _ASC)
 
     def _desc(self):
-        return SimpleOrdering(self, DESC)
+        return SimpleOrdering(self, _DESC)
+
+    desc = descending = property(_desc)
+    asc = ascending = property(_asc)
+
+
+
+class _ContainableMixin:
+    def oneOf(self, seq, negate=False):
+        """
+        Choose items whose attributes are in a fixed set.
+
+        X.oneOf([1, 2, 3])
+
+        Implemented with the SQL 'in' statement.
+        """
+        return SequenceComparison(self, seq, negate)
+
+
+    def notOneOf(self, seq):
+        return self.oneOf(seq, negate=True)
+
+
+
+class Comparable(_ContainableMixin, _ComparisonOperatorMuxer,
+                 _MatchingOperationMuxer, _OrderingMixin):
+    """
+    Helper for a thing that can be compared like an SQLAttribute (or is in fact
+    an SQLAttribute).  Requires that 'self' have 'type' (Item-subclass) and
+    'columnName' (str) attributes, as well as an 'infilter' method in the
+    spirit of SQLAttribute, documented below.
+    """
+
+    # XXX TODO: improve error reporting
+
+    def compare(self, other, sqlop):
+        return compare(self, other, sqlop)
 
-    descending = property(_desc)
-    ascending = property(_asc)
-    asc = ascending
-    desc = descending
 
-ASC = 'ASC'
-DESC = 'DESC'
 
 class SimpleOrdering:
     """
@@ -150,25 +189,21 @@ class SimpleOrdering:
     # maybe this will be a useful public API, for the query something
     # something.
 
-    isDescending = property(lambda self: self.direction == DESC)
-    isAscending = property(lambda self: self.direction == ASC)
+    isDescending = property(lambda self: self.direction == _DESC)
+    isAscending = property(lambda self: self.direction == _ASC)
 
     def __init__(self, attribute, direction=''):
         self.attribute = attribute
         self.direction = direction
 
-    def columnAndDirection(self, store):
-        """
-        'Internal' API.  Called by CompoundOrdering.
-        """
-        return '%s %s' % (self.attribute.getColumnName(store),
-                          self.direction)
 
-    def orderSQL(self, store):
-        """
-        'External' API.  Called by Query objects in store.py.
-        """
-        return 'ORDER BY ' + self.columnAndDirection(store)
+    def orderColumns(self):
+        return [(self.attribute, self.direction)]
+
+
+    def __repr__(self):
+        return repr(self.attribute) + self.direction
+
 
     def __add__(self, other):
         if isinstance(other, SimpleOrdering):
@@ -178,6 +213,7 @@ class SimpleOrdering:
         else:
             return NotImplemented
 
+
     def __radd__(self, other):
         if isinstance(other, SimpleOrdering):
             return CompoundOrdering([other, self])
@@ -196,6 +232,11 @@ class CompoundOrdering:
     def __init__(self, seq):
         self.simpleOrderings = list(seq)
 
+
+    def __repr__(self):
+        return self.__class__.__name__ + '(' + repr(self.simpleOrderings) + ')'
+
+
     def __add__(self, other):
         """
         Just thinking about what might be useful from the perspective of
@@ -211,6 +252,7 @@ class CompoundOrdering:
         else:
             return NotImplemented
 
+
     def __radd__(self, other):
         """
         Just thinking about what might be useful from the perspective of
@@ -226,8 +268,14 @@ class CompoundOrdering:
         else:
             return NotImplemented
 
-    def orderSQL(self, store):
-        return 'ORDER BY ' + (', '.join([o.columnAndDirection(store) for o in self.simpleOrderings]))
+
+    def orderColumns(self):
+        x = []
+        for o in self.simpleOrderings:
+            x.extend(o.orderColumns())
+        return x
+
+
 
 class UnspecifiedOrdering:
     implements(IOrdering)
@@ -240,8 +288,9 @@ class UnspecifiedOrdering:
 
     __radd__ = __add__
 
-    def orderSQL(self, store):
-        return ''
+
+    def orderColumns(self):
+        return []
 
 
 registerAdapter(CompoundOrdering, list, IOrdering)
@@ -264,6 +313,8 @@ class SQLAttribute(inmemory, Comparable):
 
     @ivar default: The value used for this attribute, if no value is specified.
     """
+    implements(IColumn)
+
     sqltype = None
 
     def __init__(self, doc='', indexed=False, default=None, allowNone=True, defaultFactory=None):
@@ -277,18 +328,24 @@ class SQLAttribute(inmemory, Comparable):
             raise ValueError("You may specify only one of default "
                              "or defaultFactory, not both")
 
-    def getColumnName(self, st):
-        return self.type.getTableName(st) + '.' + self.columnName
-
     def computeDefault(self):
         if self.defaultFactory is not None:
             return self.defaultFactory()
         return self.default
 
+
     def reprFor(self, oself):
         return repr(self.__get__(oself))
 
 
+    def getShortColumnName(self, store):
+        return store.getShortColumnName(self)
+
+
+    def getColumnName(self, store):
+        return store.getColumnName(self)
+
+
     def prepareInsert(self, oself, store):
         """
         Override this method to do something to an item to prepare for its
@@ -330,7 +387,6 @@ class SQLAttribute(inmemory, Comparable):
         self.modname = modname
         self.classname = classname
         self.attrname = attrname
-        self.columnName = '['+attrname+']'
         self.underlying = self.prefix + attrname
         self.dbunderlying = self.dbprefix + attrname
         yield self.underlying
@@ -404,19 +460,33 @@ class SQLAttribute(inmemory, Comparable):
                                         # attribute errors; what gives?  good
                                         # for us, I guess.
 
-    def __set__(self, oself, pyval):
+    def _convertPyval(self, oself, pyval):
+        """
+        Convert a Python value to a value suitable for inserting into the database.
+        @param oself: The object on which this descriptor is an attribute.
+        @param pyval: The value to be converted.
+        @return: A value legal for this column in the database.
+        """
         # convert to dbval later, I guess?
         if pyval is None and not self.allowNone:
             raise TypeError("attribute [%s.%s = %s()] must not be None" % (
                     self.classname, self.attrname, self.__class__.__name__))
 
+        return self.infilter(pyval, oself, oself.store)
+
+    def __set__(self, oself, pyval):
         st = oself.store
-        dbval = self.infilter(pyval, oself, st)
+
+        dbval = self._convertPyval(oself, pyval)
         oself.__dirty__[self.attrname] = self, dbval
         oself.touch()
         setattr(oself, self.underlying, pyval)
         if st is not None and st.autocommit:
-            oself.checkpoint()
+            st._rejectChanges += 1
+            try:
+                oself.checkpoint()
+            finally:
+                st._rejectChanges -= 1
 
 
 class TwoAttributeComparison:
@@ -429,16 +499,26 @@ class TwoAttributeComparison:
     def getQuery(self, store):
         sql = ('(%s %s %s)' % (self.leftAttribute.getColumnName(store),
                                self.operationString,
-                               self.rightAttribute.getColumnName(store)))
+                               self.rightAttribute.getColumnName(store)) )
         return sql
 
     def getInvolvedTables(self):
-        return set([self.leftAttribute.type, self.rightAttribute.type])
+        tables = [self.leftAttribute.type]
+        if self.leftAttribute.type is not self.rightAttribute.type:
+            tables.append(self.rightAttribute.type)
+        return tables
+
 
     def getArgs(self, store):
         return []
 
 
+    def __repr__(self):
+        return ' '.join((self.leftAttribute.fullyQualifiedName(),
+                         self.operationString,
+                         self.rightAttribute.fullyQualifiedName()))
+
+
 class AttributeValueComparison:
     implements(IComparison)
     def __init__(self, attribute, operationString, value):
@@ -454,7 +534,7 @@ class AttributeValueComparison:
         return [self.attribute.infilter(self.value, None, store)]
 
     def getInvolvedTables(self):
-        return set([self.attribute.type])
+        return [self.attribute.type]
 
     def __repr__(self):
         return ' '.join((self.attribute.fullyQualifiedName(),
@@ -479,7 +559,7 @@ class NullComparison:
         return []
 
     def getInvolvedTables(self):
-        return set([self.attribute.type])
+        return [self.attribute.type]
 
 class LikeFragment:
     def getLikeArgs(self):
@@ -527,10 +607,11 @@ class LikeComparison:
         self.likeParts = likeParts
 
     def getInvolvedTables(self):
-        tbls = set()
+        tables = [self.attribute.type]
         for lf in self.likeParts:
-            tbls.update(lf.getLikeTables())
-        return tbls
+            tables.extend([
+                    t for t in lf.getLikeTables() if t not in tables])
+        return tables
 
     def getQuery(self, store):
         if self.negate:
@@ -583,34 +664,110 @@ class AggregateComparison:
         return args
 
     def getInvolvedTables(self):
-        tbls = set()
-        # We only want to join these tables ONCE per expression
-        # OR(A.foo=='bar', A.foo=='shoe') should not do "FROM foo, foo"
+        tables = []
         for cond in self.conditions:
-            tbls.update(cond.getInvolvedTables())
-        return tbls
+            tables.extend([
+                    t for t in cond.getInvolvedTables() if t not in tables])
+        return tables
 
     def __repr__(self):
         return '%s(%s)' % (self.__class__.__name__,
                            ', '.join(map(repr, self.conditions)))
 
+
+
 class SequenceComparison:
-    def __init__(self, attribute, seq, negate):
+    implements(IComparison)
+
+    def __init__(self, attribute, container, negate):
         self.attribute = attribute
-        self.seq = seq
+        self.container = container
         self.negate = negate
 
+        if IColumn.providedBy(container):
+            self.containerClause = self._columnContainer
+            self.getArgs = self._columnArgs
+        elif IQuery.providedBy(container):
+            self.containerClause = self._queryContainer
+            self.getArgs = self._queryArgs
+        else:
+            self.containerClause = self._sequenceContainer
+            self.getArgs = self._sequenceArgs
+
+
+    def _columnContainer(self, store):
+        """
+        Return the fully qualified name of the column being examined so as
+        to push all of the containment testing into the database.
+        """
+        return self.container.getColumnName(store)
+
+
+    def _columnArgs(self, store):
+        """
+        The IColumn form of this has no arguments, just a column name
+        specified in the SQL, specified by _columnContainer.
+        """
+        return []
+
+
+    _subselectSQL = None
+    _subselectArgs = None
+    def _queryContainer(self, store):
+        """
+        Generate and cache the subselect SQL and its arguments.  Return the
+        subselect SQL.
+        """
+        if self._subselectSQL is None:
+            sql, args = self.container._sqlAndArgs('SELECT',
+                                                   self.container._queryTarget)
+            self._subselectSQL, self._subselectArgs = sql, args
+        return self._subselectSQL
+
+
+    def _queryArgs(self, store):
+        """
+        Make sure subselect arguments have been generated and then return
+        them.
+        """
+        self._queryContainer(store)
+        return self._subselectArgs
+
+
+    _sequence = None
+    def _sequenceContainer(self, store):
+        """
+        Smash whatever we got into a list and save the result in case we are
+        executed multiple times.  This keeps us from tripping up over
+        generators and the like.
+        """
+        if self._sequence is None:
+            self._sequence = list(self.container)
+            self._clause = ', '.join(['?'] * len(self._sequence))
+        return self._clause
+
+
+    def _sequenceArgs(self, store):
+        """
+        Filter each element of the data using the attribute type being
+        tested for containment and hand back the resulting list.
+        """
+        self._sequenceContainer(store) # Force _sequence to be valid
+        return [self.attribute.infilter(pyval, None, store) for pyval in self._sequence]
+
+
+    # IComparison - getArgs is assigned as an instance attribute
     def getQuery(self, store):
         return '%s %sIN (%s)' % (
             self.attribute.getColumnName(store),
             self.negate and 'NOT ' or '',
-            ', '.join(['?'] * len(self.seq)))
+            self.containerClause(store))
 
-    def getArgs(self, store):
-        return [self.attribute.infilter(pyval, None, store) for pyval in self.seq]
 
     def getInvolvedTables(self):
-        return set([self.attribute.type])
+        return [self.attribute.type]
+
+
 
 class AND(AggregateComparison):
     """
@@ -624,6 +781,37 @@ class OR(AggregateComparison):
     """
     operator = 'OR'
 
+
+class TableOrderComparisonWrapper(object):
+    """
+    Wrap any other L{IComparison} and override its L{getInvolvedTables} method
+    to specify the same tables but in an explicitly specified order.
+    """
+    implements(IComparison)
+
+    tables = None
+    comparison = None
+
+    def __init__(self, tables, comparison):
+        assert set(tables) == set(comparison.getInvolvedTables())
+
+        self.tables = tables
+        self.comparison = comparison
+
+
+    def getInvolvedTables(self):
+        return self.tables
+
+
+    def getQuery(self, store):
+        return self.comparison.getQuery(store)
+
+
+    def getArgs(self, store):
+        return self.comparison.getArgs(store)
+
+
+
 class boolean(SQLAttribute):
     sqltype = 'BOOLEAN'
 
@@ -752,18 +940,27 @@ class text(SQLAttribute):
 class textlist(text):
     delimiter = u'\u001f'
 
+    # Once upon a time, textlist encoded the list in such a way that caused []
+    # to be indistinguishable from [u'']. This value is now used as a
+    # placeholder at the head of the list, to avoid this problem in a way that
+    # is almost completely backwards-compatible with older databases.
+    guard = u'\u0002'
+
     def outfilter(self, dbval, oself):
         unicodeString = super(textlist, self).outfilter(dbval, oself)
         if unicodeString is None:
             return None
-        return unicodeString.split(self.delimiter)
+        val = unicodeString.split(self.delimiter)
+        if val[:1] == [self.guard]:
+            del val[:1]
+        return val
 
     def infilter(self, pyval, oself, store):
         if pyval is None:
             return None
         for innerVal in pyval:
-            assert self.delimiter not in innerVal
-        result = self.delimiter.join(pyval)
+            assert self.delimiter not in innerVal and self.guard not in innerVal
+        result = self.delimiter.join([self.guard] + list(pyval))
         return super(textlist, self).infilter(result, oself, store)
 
 class path(text):
@@ -852,6 +1049,7 @@ class timestamp(integer):
         return Time.fromPOSIXTimestamp(dbval / MICRO)
 
 _cascadingDeletes = {}
+_disallows = {}
 
 class reference(integer):
     NULLIFY = object()
@@ -872,11 +1070,16 @@ class reference(integer):
             # Note; this list is technically in a slightly inconsistent state
             # as things are being built.
             _cascadingDeletes.setdefault(reftype, []).append(self)
+        if whenDeleted is reference.DISALLOW:
+            _disallows.setdefault(reftype, []).append(self)
 
     def reprFor(self, oself):
         obj = getattr(oself, self.underlying, None)
         if obj is not None:
-            return 'reference(%d)' % (obj.storeID,)
+            if obj.storeID is not None:
+                return 'reference(%d)' % (obj.storeID,)
+            else:
+                return 'reference(unstored@%d)' % (id(obj),)
         sid = getattr(oself, self.dbunderlying, None)
         if sid is None:
             return 'None'
@@ -927,7 +1130,11 @@ class reference(integer):
     def outfilter(self, dbval, oself):
         if dbval is None:
             return None
-        return oself.store.getItemByID(dbval, autoUpgrade=not oself.__legacy__)
+
+        referee = oself.store.getItemByID(dbval, default=None, autoUpgrade=not oself.__legacy__)
+        if referee is None and self.whenDeleted is not reference.NULLIFY:
+            raise BrokenReference('Reference to storeID %r is broken' % (dbval,))
+        return referee
 
 class ieee754_double(SQLAttribute):
     """
