@@ -1,17 +1,25 @@
 # -*- test-case-name: axiom.test -*-
+
 __metaclass__ = type
 
+from zope.interface import implements
+
 from inspect import getabsfile
 
 from twisted.python import log
 from twisted.python.reflect import qual, namedAny
 from twisted.python.util import unsignedID
+from twisted.python.util import mergeFunctionMetadata
 from twisted.application.service import IService, IServiceCollection, MultiService
 
 from axiom import slotmachine, _schema, iaxiom
+from axiom.errors import ChangeRejected, DeletionDisallowed
+from axiom.iaxiom import IColumn, IPowerupIndirector
 
-from axiom.attributes import SQLAttribute, Comparable, inmemory, \
-    reference, text, integer, AND, _cascadingDeletes
+from axiom.attributes import (
+    SQLAttribute, _ComparisonOperatorMuxer, _MatchingOperationMuxer,
+    _OrderingMixin, _ContainableMixin, Comparable, compare, inmemory,
+    reference, text, integer, AND, _cascadingDeletes, _disallows)
 
 _typeNameToMostRecentClass = {}
 
@@ -74,11 +82,17 @@ class MetaItem(slotmachine.SchemaMetaMachine):
         _typeNameToMostRecentClass[T.typeName] = T
         return T
 
-def TABLE_NAME(st, typename, version):
-    return "%s.item_%s_v%d" % (st.databaseName, typename, version)
 
-def INDEX_NAME(st, typename, version, attrname):
-    return "%s.axiomidx_%s_v%d_%s" % (st.databaseName, typename, version, attrname)
+    def __cmp__(self, other):
+        """
+        Ensure stable sorting between Item classes.  This provides determinism
+        in SQL generation, which is beneficial for debugging and performance
+        purposes.
+        """
+        if isinstance(other, MetaItem):
+            return cmp(self.typeName, other.typeName)
+        return NotImplemented
+
 
 def noop():
     pass
@@ -87,13 +101,21 @@ class _StoreIDComparer(Comparable):
     """
     See Comparable's docstring for the explanation of the requirements of my implementation.
     """
-    columnName = 'oid'
+    implements(IColumn)
 
     def __init__(self, type):
         self.type = type
 
-    def getColumnName(self, st):
-        return self.type.getTableName(st) + '.oid'
+    def __repr__(self):
+        return '<storeID ' + qual(self.type) + '.storeID>'
+
+    def fullyQualifiedName(self):
+        # XXX: this is an example of silly redundancy, this really ought to be
+        # refactored to work like any other attribute (including being
+        # explicitly covered in the schema, which has other good qualities like
+        # allowing tables to be VACUUM'd without destroying oid stability and
+        # every storeID reference ever. --glyph
+        return qual(self.type)+'.storeID'
 
     # attributes required by ColumnComparer
     def infilter(self, pyval, oself, store):
@@ -102,6 +124,18 @@ class _StoreIDComparer(Comparable):
     def outfilter(self, dbval, oself):
         return dbval
 
+    def getShortColumnName(self, store):
+        return store.getShortColumnName(self)
+
+    def getColumnName(self, store):
+        return store.getColumnName(self)
+
+    def __get__(self, item, type=None):
+        if item is None:
+            return self
+        else:
+            return getattr(item, 'storeID')
+
 
 class _SpecialStoreIDAttribute(slotmachine.SetOnce):
     """
@@ -115,7 +149,11 @@ class _SpecialStoreIDAttribute(slotmachine.SetOnce):
     """
     def __get__(self, oself, type=None):
         if type is not None and oself is None:
-            return _StoreIDComparer(type)
+            if type._storeIDComparer is None:
+                # Reuse the same instance so that the store can use it
+                # as a key for various caching, like any other attributes.
+                type._storeIDComparer = _StoreIDComparer(type)
+            return type._storeIDComparer
         return super(_SpecialStoreIDAttribute, self).__get__(oself, type)
 
 
@@ -130,21 +168,6 @@ def serviceSpecialCase(item, pups):
 
 
 
-class InstallableMixin(object):
-    """
-    Mixin for Items which can be installed as powerups on other Items
-    and which must keep track of which item they are installed on.
-
-    If a classes mixing this in defines named C{installedOn}, it will
-    be bound to the object which is the target of the install.
-    """
-    def installOn(self, other):
-        if hasattr(self, 'installedOn'):
-            assert self.installedOn is None or self.installedOn is other, (
-                "Cannot install %r on more than one thing (was installed on "
-                "%r, tried to install on %r)" % (self.__class__, self.installedOn, other))
-            self.installedOn = other
-
 
 class Empowered(object):
 
@@ -176,7 +199,16 @@ class Empowered(object):
 
         @param priority: An int; preferably either POWERUP_BEFORE,
         POWERUP_AFTER, or unspecified.
+
+        @raise TypeError: raises if interface is IPowerupIndirector You may not
+        install a powerup for IPowerupIndirector because that would be
+        nonsensical.
         """
+        if interface is IPowerupIndirector:
+            raise TypeError(
+                "You cannot install a powerup for IPowerupIndirector: " +
+                powerup)
+
         forc = self.store.findOrCreate(_PowerupConnector,
                                        item=self,
                                        interface=unicode(qual(interface)),
@@ -207,7 +239,12 @@ class Empowered(object):
         special rules.  The full list of such interfaces is present in the
         'aggregateInterfaces' class attribute.
         """
-        if not self.store:
+        if (self.store is None  # Don't bother doing a *query* if we're not
+                                # even stored in a store yet
+            or interface is IPowerupIndirector): # you can't do a query for
+                                                 # IPowerupIndirector, that
+                                                 # would just start an infinite
+                                                 # loop.
             return
         pups = self.powerupsFor(interface)
         agg = self.aggregateInterfaces
@@ -220,6 +257,10 @@ class Empowered(object):
         """
         Returns powerups installed using C{powerUp}, in order of descending
         priority.
+
+        Powerups found to have been deleted, either during the course of this
+        powerupsFor iteration, during an upgrader, or previously, will not be
+        returned.
         """
         name = unicode(qual(interface), 'ascii')
         for cable in self.store.query(
@@ -227,7 +268,16 @@ class Empowered(object):
             AND(_PowerupConnector.interface == name,
                 _PowerupConnector.item == self),
             sort=_PowerupConnector.priority.descending):
-            yield cable.powerup
+            pup = cable.powerup
+            if pup is None:
+                # this powerup was probably deleted during an upgrader.
+                cable.deleteFromStore()
+            else:
+                indirector = IPowerupIndirector(pup, None)
+                if indirector is not None:
+                    yield indirector.indirect(interface)
+                else:
+                    yield pup
 
     def interfacesFor(self, powerup):
         pc = _PowerupConnector
@@ -236,11 +286,67 @@ class Empowered(object):
                                           pc.powerup == powerup)).getColumn('interface'):
             yield namedAny(iface)
 
-def transacted(callable):
-    def _(item, *a, **kw):
-        return item.store.transact(callable, item, *a, **kw)
-    _.func_name = callable.func_name
-    return _
+
+def transacted(func):
+    """
+    Return a callable which will invoke C{func} in a transaction using the
+    C{store} attribute of the first parameter passed to it.  Typically this is
+    used to create Item methods which are automatically run in a transaction.
+
+    The attributes of the returned callable will resemble those of C{func} as
+    closely as L{twisted.python.util.mergeFunctionMetadata} can make them.
+    """
+    def transactionified(item, *a, **kw):
+        return item.store.transact(func, item, *a, **kw)
+    return mergeFunctionMetadata(func, transactionified)
+
+
+
+
+def dependentItems(store, tableClass, comparisonFactory):
+    """
+    Collect all the items that should be deleted when an item or items
+    of a particular item type are deleted.
+
+    @param tableClass: An L{Item} subclass.
+
+    @param comparison: A one-argument callable taking an attribute and
+    returning an L{iaxiom.IComparison} describing the items to
+    collect.
+
+    @return: An iterable of items to delete.
+    """
+    for cascadingAttr in (_cascadingDeletes.get(tableClass, []) +
+                          _cascadingDeletes.get(None, [])):
+        for cascadedItem in store.query(cascadingAttr.type,
+                                        comparisonFactory(cascadingAttr)):
+            yield cascadedItem
+
+
+
+def allowDeletion(store, tableClass, comparisonFactory):
+    """
+    Returns a C{bool} indicating whether deletion of an item or items of a
+    particular item type should be allowed to proceed.
+
+    @param tableClass: An L{Item} subclass.
+
+    @param comparison: A one-argument callable taking an attribute and
+    returning an L{iaxiom.IComparison} describing the items to
+    collect.
+
+    @return: A C{bool} indicating whether deletion should be allowed.
+    """
+    for cascadingAttr in (_disallows.get(tableClass, []) +
+                          _disallows.get(None, [])):
+        for cascadedItem in store.query(cascadingAttr.type,
+                                        comparisonFactory(cascadingAttr),
+                                        limit=1):
+            return False
+
+    return True
+
+
 
 class Item(Empowered, slotmachine._Strict):
     # Python-Special Attributes
@@ -272,6 +378,7 @@ class Item(Empowered, slotmachine._Strict):
                                   # part of an upgrade)
 
     storeID = _SpecialStoreIDAttribute(default=None)
+    _storeIDComparer = None
     _axiom_service = inmemory()
 
     def _currentlyValidAsReferentFor(self, store):
@@ -280,12 +387,27 @@ class Item(Empowered, slotmachine._Strict):
         deleted in this transaction, or objects which are not in the same store
         are not valid.  See attributes.reference.__get__.
         """
+        if store is None:
+            # If your store is None, you can refer to whoever you want.  I'm in
+            # a store but it doesn't matter that you're not.
+            return True
         if self.store is not store:
             return False
         if self.__deletingObject:
             return False
         return True
-    # return (self.store is store and not self.__deletingObject)
+
+
+    def _schemaPrepareInsert(self, store):
+        """
+        Prepare each attribute in my schema for insertion into a given store,
+        either by upgrade or by creation.  This makes sure all references point
+        to this store and all relative paths point to this store's files
+        directory.
+        """
+        for name, atr in self.getSchema():
+            atr.prepareInsert(self, store)
+
 
     def store():
         def get(self):
@@ -294,9 +416,11 @@ class Item(Empowered, slotmachine._Strict):
             if self.__store is not None:
                 raise AttributeError(
                     "Store already set - can't move between stores")
-            # make damn sure all of our references point to this store:
-            for name, atr in self.getSchema():
-                atr.prepareInsert(self, store)
+
+            if store._rejectChanges:
+                raise ChangeRejected()
+
+            self._schemaPrepareInsert(store)
             self.__store = store
             oid = self.storeID = self.store.executeSchemaSQL(
                 _schema.CREATE_OBJECT, [self.store.getTypeID(type(self))])
@@ -309,6 +433,7 @@ class Item(Empowered, slotmachine._Strict):
             else:
                 self.touch()
             self.activate()
+            self.stored()
         return get, set, """
 
         A reference to a Store; when set for the first time, inserts this object
@@ -320,9 +445,12 @@ class Item(Empowered, slotmachine._Strict):
 
     store = property(*store())
 
-# XXX: Think about how to do this _safely_ (e.g. not recursing infinitely
-# through circular references) before turning it on
     def __repr__(self):
+        """
+        Return a nice string representation of the Item which contains some
+        information about each of its attributes.
+        """
+
         L = [self.__name__]
         L.append('(')
         A = []
@@ -368,8 +496,14 @@ class Item(Empowered, slotmachine._Strict):
 
         if tostore != None:
             if to__justUpgraded:
+
+                # we can't just set the store, because that allocates an ID.
+                # we do still need to do all the attribute prep, make sure
+                # references refer to this store, paths are adjusted to point
+                # to this store's static offset, etc.
+
+                self._schemaPrepareInsert(tostore)
                 self.__store = tostore
-                # skip all the interesting just-created stuff, like ID allocation.
                 if tostore.autocommit:
                     self.checkpoint()
             else:
@@ -439,9 +573,9 @@ class Item(Empowered, slotmachine._Strict):
 
     def touch(self):
         # xxx what
-        if self.store is None or self.store.transaction is None:
+        if self.store is None:
             return
-        self.store.transaction.add(self)
+        self.store.changed(self)
 
     def revert(self):
         if self.__justCreated:
@@ -451,8 +585,7 @@ class Item(Empowered, slotmachine._Strict):
             return
         self.__dirty__.clear()
         dbattrs = self.store.querySQL(
-            self.store.getTableQuery(self.typeName,
-                                     self.schemaVersion),
+            self._baseSelectSQL(self.store),
             [self.storeID])[0]
 
         for data, (name, atr) in zip(dbattrs, self.getSchema()):
@@ -461,12 +594,24 @@ class Item(Empowered, slotmachine._Strict):
         self.__deleting = False
         self.__deletingObject = False
 
+
     def deleted(self):
         """User-definable callback that is invoked when an object is well and truly
         gone from the database; the transaction which deleted it has been
         committed.
         """
 
+
+    def stored(self):
+        """
+        User-definable callback that is invoked when an object is placed into a
+        Store for the very first time.
+
+        If an Item is created with a store, this will be invoked I{after}
+        C{activate}.
+        """
+
+
     def committed(self):
         """
         Called after the database is brought into a consistent state with this
@@ -479,8 +624,10 @@ class Item(Empowered, slotmachine._Strict):
             self.__store = None
         self.__justCreated = False
 
+
     def checkpoint(self):
-        """ Update the database to reflect in-memory changes made to this item; for
+        """
+        Update the database to reflect in-memory changes made to this item; for
         example, to make it show up in store.query() calls where it is now
         valid, but was not the last time it was persisted to the database.
 
@@ -559,7 +706,6 @@ class Item(Empowered, slotmachine._Strict):
         # functioning as expected
         assert typename == self.typeName, '%r != %r' % (typename, self.typeName)
         assert oldversion == self.schemaVersion
-        assert newversion == oldversion + 1
         key = typename, newversion
         T = None
         if key in _legacyTypes:
@@ -592,18 +738,17 @@ class Item(Empowered, slotmachine._Strict):
         self.deleteFromStore(False)
         return new
 
-    def dependentItems(self):
-        for cascadingAttr in (_cascadingDeletes.get(self.__class__, []) +
-                              _cascadingDeletes.get(None, [])):
-            for cascadedItem in self.store.query(cascadingAttr.type,
-                                                 cascadingAttr == self):
-                yield cascadedItem
-
-
     def deleteFromStore(self, deleteObject=True):
         # go grab dependent stuff
         if deleteObject:
-            for dependent in self.dependentItems():
+            if not allowDeletion(self.store, self.__class__,
+                                 lambda attr: attr == self):
+                raise DeletionDisallowed(
+                    'Cannot delete item; '
+                    'has referents with whenDeleted == reference.DISALLOW')
+
+            for dependent in dependentItems(self.store, self.__class__,
+                                            lambda attr: attr == self):
                 dependent.deleteFromStore()
 
         self.touch()
@@ -621,16 +766,17 @@ class Item(Empowered, slotmachine._Strict):
 
     ###### SQL generation ######
 
-    def getTableName(cls, st):
-        return TABLE_NAME(st, cls.typeName, cls.schemaVersion)
+    def _baseSelectSQL(cls, st):
+        if cls not in st.typeToSelectSQLCache:
+            st.typeToSelectSQLCache[cls] = ' '.join(['SELECT * FROM',
+                                                     st.getTableName(cls),
+                                                     'WHERE',
+                                                     st.getShortColumnName(cls.storeID),
+                                                     '= ?'
+                                                     ])
+        return st.typeToSelectSQLCache[cls]
 
-    getTableName = classmethod(getTableName)
-
-    def indexNameOf(cls, st, attrsnames):
-        "private to store"
-        return INDEX_NAME(st, cls.typeName, cls.schemaVersion, attrsnames)
-
-    indexNameOf = classmethod(indexNameOf)
+    _baseSelectSQL = classmethod(_baseSelectSQL)
 
     def _baseInsertSQL(cls, st):
         if cls not in st.typeToInsertSQLCache:
@@ -638,9 +784,9 @@ class Item(Empowered, slotmachine._Strict):
             qs = ', '.join((['?']*(len(attrs)+1)))
             st.typeToInsertSQLCache[cls] = (
                 'INSERT INTO '+
-                cls.getTableName(st) + ' (' + ', '.join(
-                    ['oid'] +
-                    [a[1].columnName for a in attrs]) +
+                st.getTableName(cls) + ' (' + ', '.join(
+                    [ st.getShortColumnName(cls.storeID) ] +
+                    [ st.getShortColumnName(a[1]) for a in attrs]) +
                 ') VALUES (' + qs + ')')
         return st.typeToInsertSQLCache[cls]
 
@@ -649,8 +795,10 @@ class Item(Empowered, slotmachine._Strict):
     def _baseDeleteSQL(cls, st):
         if cls not in st.typeToDeleteSQLCache:
             st.typeToDeleteSQLCache[cls] = ' '.join(['DELETE FROM',
-                                                     cls.getTableName(st),
-                                                     'WHERE oid = ? '
+                                                     st.getTableName(cls),
+                                                     'WHERE',
+                                                     st.getShortColumnName(cls.storeID),
+                                                     '= ? '
                                                      ])
         return st.typeToDeleteSQLCache[cls]
 
@@ -667,17 +815,160 @@ class Item(Empowered, slotmachine._Strict):
         dirtyColumns = []
         dirtyValues = []
         for dirtyAttrName, (dirtyAttribute, dirtyValue) in dirty:
-            dirtyColumns.append(dirtyAttribute.columnName)
+            dirtyColumns.append(self.store.getShortColumnName(dirtyAttribute))
             dirtyValues.append(dirtyValue)
         stmt = ' '.join([
-            'UPDATE', self.getTableName(self.store), 'SET',
+            'UPDATE', self.store.getTableName(self.__class__), 'SET',
              ', '.join(['%s = ?'] * len(dirty)) %
               tuple(dirtyColumns),
-              'WHERE oid = ?'])
+            'WHERE ', self.store.getShortColumnName(type(self).storeID), ' = ?'])
         dirtyValues.append(self.storeID)
         return stmt, dirtyValues
 
 
+    def getTableName(cls, store):
+        """
+        Retrieve a string naming the database table associated with this item
+        class.
+        """
+        return store.getTableName(cls)
+    getTableName = classmethod(getTableName)
+
+
+    def getTableAlias(cls, store, currentAliases):
+        return None
+    getTableAlias = classmethod(getTableAlias)
+
+
+
+class _PlaceholderColumn(_ContainableMixin, _ComparisonOperatorMuxer,
+                         _MatchingOperationMuxer, _OrderingMixin):
+    """
+    Wrapper for columns from a L{Placeholder} which provides a fully qualified
+    name built with a table alias name instead of the underlying column's real
+    table name.
+    """
+    implements(IColumn)
+
+    def __init__(self, placeholder, column):
+        self.type = placeholder
+        self.column = column
+
+    def __repr__(self):
+        return '<Placeholder %r>' % (self.column,)
+
+
+    def __get__(self, inst):
+        return self.column.__get__(inst)
+
+    def fullyQualifiedName(self):
+        return self.column.fullyQualifiedName() + '.<placeholder:%s>' % (
+            self.type._placeholderCount,)
+
+
+    def compare(self, other, op):
+        return compare(self, other, op)
+
+
+    def getShortColumnName(self, store):
+        return self.column.getShortColumnName(store)
+
+
+    def getColumnName(self, store):
+        assert self.type._placeholderTableAlias is not None, (
+            "Placeholder.getTableAlias() must be called "
+            "before Placeholder.attribute.getColumnName()")
+
+        return '%s.%s' % (self.type._placeholderTableAlias,
+                          self.column.getShortColumnName(store))
+
+    def infilter(self, pyval, oself, store):
+        return self.column.infilter(pyval, oself, store)
+
+
+    def outfilter(self, dbval, oself):
+        return self.column.outfilter(dbval, oself)
+
+
+
+_placeholderCount = 0
+
+class Placeholder(object):
+    """
+    Wrap an existing L{Item} type to provide a different name for it.
+
+    This can be used to join a table against itself which is useful for
+    flattening normalized data.  For example, given a schema defined like
+    this::
+
+        class Tag(Item):
+            taggedObject = reference()
+            tagName = text()
+
+
+        class SomethingElse(Item):
+            ...
+
+
+    It might be useful to construct a query for instances of SomethingElse
+    which have been tagged both with C{"foo"} and C{"bar"}::
+
+        t1 = Placeholder(Tag)
+        t2 = Placeholder(Tag)
+        store.query(SomethingElse, AND(t1.taggedObject == SomethingElse.storeID,
+                                       t1.tagName == u"foo",
+                                       t2.taggedObject == SomethingElse.storeID,
+                                       t2.tagName == u"bar"))
+    """
+    _placeholderTableAlias = None
+
+    def __init__(self, itemClass):
+        global _placeholderCount
+
+        self._placeholderItemClass = itemClass
+        self._placeholderCount = _placeholderCount + 1
+        _placeholderCount += 1
+
+        self.existingInStore = self._placeholderItemClass.existingInStore
+
+
+    def __cmp__(self, other):
+        """
+        Provide a deterministic sort order between Placeholder instances.
+        Those instantiated first will compare as less than than instantiated
+        later.
+        """
+        if isinstance(other, Placeholder):
+            return cmp(self._placeholderCount, other._placeholderCount)
+        return NotImplemented
+
+
+    def __getattr__(self, name):
+        if name == 'storeID' or name in dict(self._placeholderItemClass.getSchema()):
+            return _PlaceholderColumn(self, getattr(self._placeholderItemClass, name))
+        raise AttributeError(name)
+
+
+    def getSchema(self):
+        # In a MultipleItemQuery, the same table can appear more than
+        # once in the "SELECT ..." part of the query, determined by
+        # getSchema(). In this case, the correct placeholder names
+        # need to be used.
+        for (name, atr) in self._placeholderItemClass.getSchema():
+            yield name, _PlaceholderColumn(self,
+                                           getattr(self._placeholderItemClass, name))
+
+    def getTableName(self, store):
+        return self._placeholderItemClass.getTableName(store)
+
+
+    def getTableAlias(self, store, currentAliases):
+        if self._placeholderTableAlias is None:
+            self._placeholderTableAlias = 'placeholder_' + str(len(currentAliases))
+        return self._placeholderTableAlias
+
+
+
 _legacyTypes = {}               # map (typeName, schemaVersion) to dummy class
 
 def declareLegacyItem(typeName, schemaVersion, attributes, dummyBases=()):
@@ -689,7 +980,8 @@ def declareLegacyItem(typeName, schemaVersion, attributes, dummyBases=()):
     @param typeName: a string, the Axiom TypeName to have attributes for.
     @param schemaVersion: an int, the (old) version of the schema this is a proxy
     for.
-    @param attributes: a dict mapping {columnName: attr instance}
+    @param attributes: a dict mapping {columnName: attr instance} describing
+    the schema of C{typeName} at C{schemaVersion}.
 
     @param dummyBases: a sequence of 4-tuples of (baseTypeName,
     baseSchemaVersion, baseAttributes, baseBases) representing the dummy bases
