@@ -1,4 +1,3 @@
-
 import sys
 import os
 
@@ -8,6 +7,7 @@ from twisted.python.util import sibpath
 
 from epsilon import extime
 from axiom import attributes, item, store, errors
+from axiom.errors import ItemNotFound
 
 from pysqlite2.dbapi2 import sqlite_version_info
 
@@ -19,7 +19,7 @@ class RevertException(Exception):
 class TestItem(item.Item):
     schemaVersion = 1
     typeName = 'TestItem'
-    foo = attributes.integer(indexed=True)
+    foo = attributes.integer(indexed=True, default=10)
     bar = attributes.text()
     baz = attributes.timestamp()
     other = attributes.reference()
@@ -71,6 +71,63 @@ class StoreTests(unittest.TestCase):
             s.getTableQuery(TestItem.typeName, 1))
 
 
+    def testTypeToDatabaseNames(self):
+        # The real purpose of this test is to have the new get*Name
+        # methods explicitely called somewhere in the test suite. The
+        # effect itself does not actually matter much. These functions
+        # are proven right by the fact that item creation, querying
+        # and update are working.
+
+        # I think the following should be ok for anything that vaguely
+        # ressembles SQL.
+
+        s = store.Store()
+        tn = s.getTableName(TestItem)
+
+        assert tn.startswith(s.databaseName)
+
+        cn = s.getColumnName(TestItem.foo)
+        scn = s.getShortColumnName(TestItem.foo)
+
+        assert len(tn) < len(cn)
+        assert len(scn) < len(cn)
+        assert cn.endswith(scn)
+        assert cn.startswith(tn)
+
+        icn = s.getColumnName(TestItem.storeID)
+        sicn = s.getShortColumnName(TestItem.storeID)
+
+        assert len(tn) < len(icn)
+        assert len(sicn) < len(icn)
+        assert icn.endswith(sicn)
+        assert icn.startswith(tn)
+
+    def testGetTableName(self):
+        """
+        Item instances were getting into the table name cache.
+        Make sure only classes are accepted.
+        """
+        s = store.Store()
+        self.assertRaises(errors.ItemClassesOnly, s.getTableName, TestItem(store=s))
+
+    def testTableNameCacheDoesntGrow(self):
+        """
+        Make sure the table name cache doesn't grow out of control anymore.
+        """
+        s = store.Store()
+        tn = s.getTableName(TestItem)
+        x = len(s.typeToTableNameCache)
+        for i in range(10):
+            s.getTableName(TestItem)
+        self.assertEquals(x, len(s.typeToTableNameCache))
+    def testStoreIDComparerIdentity(self):
+        # We really want this to hold, because the storeID object is
+        # used like a regular attribute as a key for various caching
+        # within store.
+        a0 = TestItem.storeID
+        a1 = TestItem.storeID
+        assert a0 is a1
+
 
 class FailurePathTests(unittest.TestCase):
 
@@ -178,6 +235,25 @@ class ItemTests(unittest.TestCase):
         self.assertRaises(KeyError, self.store.getItemByID, sid[0])
         self.assertEquals(list(self.store.query(TestItem)), [])
 
+
+    def test_getNeverInsertedItem(self):
+        """
+        Verify that using getItemByID with a default object to attempt to
+        load by storeID an Item which was created and deleted within a
+        single transaction results in the default object.
+        """
+        def txn():
+            a = TestItem(store=self.store)
+            storeID = a.storeID
+            a.deleteFromStore()
+            del a
+            return storeID
+        storeID = self.store.transact(txn)
+        default = object()
+        result = self.store.getItemByID(storeID, default=default)
+        self.assertIdentical(result, default)
+
+
     def testInMemoryRevert(self):
         item1 = TestItem(
             store=self.store,
@@ -366,6 +442,176 @@ class TestFindOrCreate(unittest.TestCase):
         self.assertEquals(a0, ai2)
 
 
+
+class DeletedTrackingItem(item.Item):
+    """
+    Helper class for testing that C{deleted} is called by
+    ItemQuery.deleteFromStore.
+    """
+    deletedTimes = 0
+    value = attributes.integer()
+
+    def deleted(self):
+        DeletedTrackingItem.deletedTimes += 1
+
+
+
+class DeleteFromStoreTrackingItem(item.Item):
+    """
+    Helper class for testing that C{deleteFromStore} is called by
+    ItemQuery.deleteFromStore.
+    """
+    deletedTimes = 0
+    value = attributes.integer()
+
+    def deleteFromStore(self):
+        DeleteFromStoreTrackingItem.deletedTimes += 1
+        item.Item.deleteFromStore(self)
+
+
+
+class MassInsertDeleteTests(unittest.TestCase):
+
+    def setUp(self):
+        self.storepath = self.mktemp()
+        self.store = store.Store(self.storepath)
+
+    def testBatchInsert(self):
+        """
+        Make sure that batchInsert creates all the items it's supposed
+        to with appropriate attributes.
+        """
+
+        dataRows = [(37, 93),
+                    (1,   2)]
+
+        self.store.batchInsert(AttributefulItem,
+                              [AttributefulItem.withDefault,
+                               AttributefulItem.withoutDefault],
+                              dataRows)
+        items = list(self.store.query(AttributefulItem))
+        self.assertEquals(items[0].withDefault, 37)
+        self.assertEquals(items[0].withoutDefault, 93)
+        self.assertEquals(items[1].withDefault, 1)
+        self.assertEquals(items[1].withoutDefault, 2)
+
+    def testTransactedBatchInsert(self):
+        """
+        Test that batchInsert works in a transaction.
+        """
+        dataRows = [(37, 93),
+                    (1,   2)]
+
+        self.store.transact(self.store.batchInsert,
+                            AttributefulItem,
+                            [AttributefulItem.withDefault,
+                             AttributefulItem.withoutDefault],
+                            dataRows)
+
+        items = list(self.store.query(AttributefulItem))
+        self.assertEquals(items[0].withDefault, 37)
+        self.assertEquals(items[0].withoutDefault, 93)
+        self.assertEquals(items[1].withDefault, 1)
+        self.assertEquals(items[1].withoutDefault, 2)
+
+    def testBatchInsertReference(self):
+        """
+        Test that reference args are handled okay by batchInsert.
+        """
+        itemA = AttributefulItem(store=self.store)
+        itemB = AttributefulItem(store=self.store)
+        dataRows = [(1, u"hello", extime.Time(),
+                     itemA, True, False, self.store),
+                    (2, u"hoorj", extime.Time(),
+                     itemB, False, True, self.store)]
+
+        self.store.batchInsert(TestItem,
+                               [TestItem.foo, TestItem.bar,
+                                TestItem.baz, TestItem.other,
+                                TestItem.booleanT, TestItem.booleanF,
+                                TestItem.myStore],
+                               dataRows)
+        items = list(self.store.query(TestItem))
+
+        self.assertEquals(items[0].other, itemA)
+        self.assertEquals(items[1].other, itemB)
+        self.assertEquals(items[0].store, self.store)
+        self.assertEquals(items[1].store, self.store)
+
+    def testMemoryBatchInsert(self):
+        """
+        Test that batchInsert works on an in-memory store.
+        """
+        self.store = store.Store()
+        self.testBatchInsert()
+
+    def testBatchInsertSelectedAttributes(self):
+        """
+        Test that batchInsert does the right thing when only a few
+        attributes are being set.
+        """
+        dataRows = [(u"hello", 50, False, self.store),
+                    (u"hoorj", None, True, self.store)]
+
+        self.store.batchInsert(TestItem,
+                               [TestItem.bar,
+                                TestItem.foo,
+                                TestItem.booleanF,
+                                TestItem.myStore],
+                               dataRows)
+        items = list(self.store.query(TestItem))
+
+        self.assertEquals(items[0].other, None)
+        self.assertEquals(items[1].other, None)
+        self.assertEquals(items[0].foo, 50)
+        self.assertEquals(items[1].foo, None)
+        self.assertEquals(items[0].bar, u"hello")
+        self.assertEquals(items[1].bar, u"hoorj")
+        self.assertEquals(items[0].store, self.store)
+        self.assertEquals(items[1].store, self.store)
+
+    def testBatchDelete(self):
+        """
+        Ensure that unqualified batchDelete removes all the items of a
+        certain class.
+        """
+        for i in xrange(10):
+            AttributefulItem(store=self.store, withoutDefault=i)
+
+        self.store.query(AttributefulItem).deleteFromStore()
+        self.assertEquals(list(self.store.query(AttributefulItem)), [])
+
+    def testBatchDeleteCondition(self):
+        """
+        Ensure that conditions for batchDelete are honored properly.
+        """
+        for i in xrange(10):
+            AttributefulItem(store=self.store, withoutDefault=i)
+
+        self.store.query(AttributefulItem,
+                              AttributefulItem.withoutDefault > 4
+                         ).deleteFromStore()
+        self.assertEquals(self.store.query(AttributefulItem).count(), 5)
+
+    def testSlowBatchDelete(self):
+        """
+        Ensure that a 'deleted' method on an Item will be called if it exists.
+        """
+        DeletedTrackingItem(store=self.store)
+        self.store.query(DeletedTrackingItem).deleteFromStore()
+        self.assertEqual(DeletedTrackingItem.deletedTimes, 1)
+
+
+    def test_slowBatchDeleteBecauseDeletedFromStore(self):
+        """
+        Ensure that a 'deleteFromStore' method on an Item will be called if it
+        exists.
+        """
+        DeleteFromStoreTrackingItem(store=self.store)
+        self.store.query(DeleteFromStoreTrackingItem).deleteFromStore()
+        self.assertEqual(DeleteFromStoreTrackingItem.deletedTimes, 1)
+
+
 # Item types we will use to change the underlying database schema (by creating
 # them).
 class ConcurrentItemA(item.Item):
@@ -460,4 +706,3 @@ class ConcurrencyTestCase(unittest.TestCase):
 
         ConcurrentItemA(store=firstStore)
         self.assertEquals(secondStore.query(ConcurrentItemA).count(), 1)
-
